/* one-directional tcp pipe tool.

Copyright (c) 2003 - 2005 Heikki Orsila <heikki.orsila@iki.fi>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>

#include "net.h"
#include "pipe.h"
#include "version.h"
#include "accounting.h"



static void print_help(int only_version)
{
  fprintf(stderr, "tcppipe %s by Heikki Orsila <heikki.orsila@iki.fi>\n", TOOL_VERSION);
  fprintf(stderr, "\n");
  if (only_version)
    return;
  fprintf(stderr, "USAGE: tcppipe [-b file] [-d file] [-r rate] [-s port] [-t port] [host port]\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "OPTIONS\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "-h                     Print help\n");
  fprintf(stderr, "-b/--bind-file file    Bind socket (listening or connecting) to an interface\n");
  fprintf(stderr, "                       determined by numbers in ip list file. The interface\n");
  fprintf(stderr, "                       with least amount of transfered bytes is selected.\n");
  fprintf(stderr, "-d/--destination-file file  Same as -b but for selecting the destination host.\n");
  fprintf(stderr, "                            Use -t to select the destination port.\n");
  fprintf(stderr, "-r/--rate rate         Maximum number of bytes per second going through one\n");
  fprintf(stderr, "                       connection.\n");
  fprintf(stderr, "-s port                Set source port to accept incoming connections\n");
  fprintf(stderr, "-t port                Set target port to connect to. Use with -d.\n");
  fprintf(stderr, "-v/--version           Print version information.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Examples:\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "  Bind to an interface on port 2345 with least amount of transfers\n");
  fprintf(stderr, "  tcppipe -s 2345 -a accountfile\n");
  fprintf(stderr, "\n");
}


int main(int argc, char *argv[])
{
  int i;
  int src_fd, dst_fd;
  char *dst_host = NULL, *dst_port = NULL;
  char *listening_port = NULL;
  char *connection_port = NULL;
  char *src_accounting_file = NULL;
  char *dst_accounting_file = NULL;

  struct pipe p1;
  char buf[1024];
  char *bind_host = NULL;
  int have_src, have_dst;

  for (i = 1; i < argc;) {

    if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
      print_help(0);
      return 0;

    } else if (!strcmp("-b", argv[i]) || !strcmp("--bind-file", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give accounting file\n");
	print_help(0);
	return -1;
      }
      src_accounting_file = strdup(argv[i + 1]);
      if (src_accounting_file == NULL) {
	fprintf(stderr, "not enough memory for accounting filename\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("-d", argv[i]) || !strcmp("--destination-file", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give accounting file\n");
	print_help(0);
	return -1;
      }
      dst_accounting_file = strdup(argv[i + 1]);
      if (dst_accounting_file == NULL) {
	fprintf(stderr, "not enough memory for accounting filename\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("-r", argv[i]) || !strcmp("--rate", argv[i])) {
      if ((i + 1) < argc) {
        pmr_max_rate = atoi(argv[i + 1]);
	if (pmr_max_rate <= 0) {
	  fprintf (stderr, "illegal bytes per second value (%d)\n", pmr_max_rate);
	  return -1;
	}
      } else {
        fprintf (stderr, "expecting a rate limit value (bytes per second) with -r\n");
	print_help(0);
        return -1;
      }
      i += 2;

    } else if (!strcmp("-s", argv[i]) || !strcmp("--source-port", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give a listening port too.\n");
	exit(-1);
      }
      listening_port = argv[i + 1];
      if (atoi(listening_port) <= 0) {
	fprintf(stderr, "illegal listening port given: %s\n", argv[i + 1]);
	return -1;
      }
      i += 2;

    } else if (!strcmp("-t", argv[i]) || !strcmp("--destination-port", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give a connection port too.\n");
	exit(-1);
      }
      connection_port = argv[i + 1];
      if (atoi(connection_port) <= 0) {
	fprintf(stderr, "illegal connection port given: %s\n", argv[i + 1]);
	return -1;
      }
      i += 2;

    } else if (!strcmp("-v", argv[i]) || !strcmp("--version", argv[i])) {
      print_help(1);
      return 0;

    } else if (argv[i][0] == '-') {
      fprintf(stderr, "illegal args (%s)\n", argv[i]);
      print_help(0);
      return -1;
    } else {
      break;
    }
  }

  if ((i + 2) == argc) {
    dst_host = argv[i];
    dst_port = argv[i + 1];
  } else if (i != argc) {
    fprintf(stderr, "illegal arguments\n");
    print_help(0);
    return -1;
  }

  if (listening_port == NULL && src_accounting_file != NULL) {
    if (dst_host == NULL) {
      fprintf(stderr, "error. a source port is needed\n");
      return -1;
    }
  }

  if (dst_host == NULL && connection_port == NULL && dst_accounting_file != NULL) {
    fprintf(stderr, "error. a target port is needed too.\n");
  }

  if (dst_host != NULL && dst_accounting_file != NULL) {
    fprintf(stderr, "error. destination specified in two ways. use only one.\n");
    return -1;
  }

  if (dst_port != NULL && connection_port != NULL) {
    fprintf(stderr, "error. destination port specified in two ways. use only one.\n");
    return -1;
  }

  have_src = (listening_port != NULL) ? 1 : 0;
  have_dst = (dst_host != NULL || dst_accounting_file != NULL) ? 1 : 0;

  if (have_src == 0 && have_dst == 0) {
    print_help(0);
    return -1;
  }

  if (have_src && have_dst) {
    fprintf(stderr, "error. will not both listen and connect. only one.\n");
    return -1;
  }

  if (src_accounting_file) {
    get_account(buf, sizeof(buf), src_accounting_file);
    bind_host = strdup(buf);
    fprintf(stderr, "binding to %s\n", bind_host);
  }

  if (dst_accounting_file) {
    if (dst_host != NULL) {
      fprintf(stderr, "error. dst host was given manually, but destination accounting file was given.\n");
      return -1;
    }
    get_account(buf, sizeof(buf), dst_accounting_file);
    dst_host = strdup(buf);
    dst_port = connection_port;
    fprintf(stderr, "connecting to %s\n", dst_host);
  }

  if (listening_port > 0) {
    int listen_fd;
    char printed_id[256];
    struct net_host_id id;
    int socklen;
    char sockbuf[256];
    struct sockaddr_in *sockname = (struct sockaddr_in *) sockbuf;

    listen_fd = net_bind(NULL, bind_host, listening_port, "tcp", 1 /* REUSE */);
    if (listen_fd < 0) {
      fprintf(stderr, "can't bind server port\n");
      return -1;
    }
    if (listen(listen_fd, 5)) {
      perror("can not listen server port");
      return -1;
    }

    while (1) {
      socklen = sizeof(sockbuf);
      src_fd = accept(listen_fd, (struct sockaddr *) sockname, &socklen);
      if (src_fd < 0) {
	if (errno == EAGAIN || errno == EINTR)
	  continue;
	perror("client accept failed");
	return -1;
      }
      break;
    }

    close(listen_fd);

    if (!net_get_id(&id, (struct sockaddr *) sockname)) {
      fprintf(stderr, "client connected. can't get info on socket. rejecting.\n");
      return -1;
    }

    printed_id[0] = 0;
    net_print_id(printed_id, &id, sizeof(printed_id));
    fprintf(stderr, "client connected from %s\n", printed_id);

    dst_fd = 1;

  } else {
    src_fd = 0;
    dst_fd = net_open(dst_host, dst_port, "tcp", bind_host);
    if (dst_fd < 0) {
      fprintf(stderr, "couldn't connect to %s:%s\n", dst_host, dst_port);
      return -1;
    }
  }

  p1 = (struct pipe) {.src_fd = src_fd, .dst_fd = dst_fd};

  pmr_shut_up = 1;

  /* here is the main loop copying data from one socket to another */
  pmr_loop(&p1, NULL, dst_host, dst_accounting_file, bind_host, src_accounting_file);

  if (src_accounting_file != NULL)
    account_increment(bind_host, src_accounting_file, p1.accbytes);
  if (dst_accounting_file != NULL)
    account_increment(dst_host, dst_accounting_file, p1.accbytes);

  close(p1.src_fd);
  close(p1.dst_fd);

  return 0;
}
